from pprint import pprint
from typing import Iterable

import pytest
import random
from negmas import *

random.seed(0)

class MyMechanism(Mechanism):
    def __init__(self, dynamic_entry=True):
        super().__init__(outcomes=20, n_steps=10
                         , dynamic_entry=dynamic_entry)
        self.current = 0

    def round(self) -> MechanismRoundResult:
        # r = random.random()
        # if r > 1.0 / self.n_steps:
        #    return None, False, None
        self.current += 1
        if self.current < 6:
            return MechanismRoundResult(agreement=None)
        return MechanismRoundResult(agreement=self.random_outcomes(1)[0])


@pytest.fixture
def mechanism():
    return MyMechanism()


@pytest.fixture
def static_mechanism():
    return MyMechanism(dynamic_entry=False)


@pytest.fixture
def dynamic_mechanism():
    return MyMechanism(dynamic_entry=True)


def test_set_requirements(mechanism):
    c = {
        'a': True,
        'b': [1, 2],
        'c': [1],
        'd': [1, 2],
        'e': False,
        'f': None,
    }
    d = {
        'g': True
    }
    mechanism.requirements = c
    # print(mechanism.requirements)
    for k, v in c.items():
        if isinstance(v, Iterable):
            assert [a == b for a, b in zip(mechanism.requirements[k] , v)]
        else:
            assert mechanism.requirements[k] == v

    mechanism.add_requirements(d)
    for k, v in c.items():
        if isinstance(v, Iterable):
            assert [a == b for a, b in zip(mechanism.requirements[k] , v)]
        else:
            assert mechanism.requirements[k] == v

    for k, v in d.items():
        if isinstance(v, Iterable):
            assert [a == b for a, b in zip(mechanism.requirements[k] , v)]
        else:
            assert mechanism.requirements[k] == v


def test_can_work_with_no_requirements(mechanism):
    assert mechanism.can_participate(RandomNegotiator(outcomes=mechanism.outcomes))


def test_can_check_compatibility(mechanism):
    c = {'a': True,
         'b': [1, 2, 3],
         'c': 1,
         'd': None,
         'f': 3, }

    assert mechanism.is_satisfying(c)

    mechanism.add_requirements({'a': False})
    assert not mechanism.is_satisfying(c)

    mechanism.add_requirements({'a': True})
    assert mechanism.is_satisfying(c)

    mechanism.add_requirements({'g': False})
    assert not mechanism.is_satisfying(c)

    mechanism.requirements = c
    assert mechanism.is_satisfying(c)

    mechanism.add_requirements({'g': False})
    assert not mechanism.is_satisfying(c)

    mechanism.remove_requirements(['a', 'g'])
    assert mechanism.is_satisfying(c)

    mechanism.requirements = c
    assert mechanism.is_satisfying(c)

    mechanism.add_requirements({'a': not c['a']})
    assert not mechanism.is_satisfying(c)

    mechanism.add_requirements({'a': None})
    assert mechanism.is_satisfying(c)

    mechanism.add_requirements({'a': not c['a']})
    assert not mechanism.is_satisfying(c)

    mechanism.add_requirements({'a': [True, False]})
    assert mechanism.is_satisfying(c)

    class HasCapabilities(object):
        capabilities = c

    assert mechanism.can_participate(HasCapabilities())


def test_different_capability_types(mechanism):
    c = {
        'bvalue': True,
        'svalue': 'abc',
        'ivalue': 1,
        'irange': (1, 5),
        'frange': (1.0, 5.0),
        'ilist': [1, 2, 9],
    }
    mechanism.requirements = c

    # single value
    assert mechanism.is_satisfying({**c, 'bvalue': True})
    assert not mechanism.is_satisfying({**c, 'bvalue': False})
    assert mechanism.is_satisfying({**c, 'bvalue': [True, False]})
    assert mechanism.is_satisfying({**c, 'svalue': 'abc'})
    assert mechanism.is_satisfying({**c, 'svalue': ['abc', 'cde']})
    assert not mechanism.is_satisfying({**c, 'svalue': 'ab'})
    assert not mechanism.is_satisfying({**c, 'svalue': ['cc','ab']})
    assert mechanism.is_satisfying({**c, 'ivalue': 1})
    assert not mechanism.is_satisfying({**c, 'ivalue': 5})

    # range
    assert not mechanism.is_satisfying({**c, 'irange': [-1, 0]})
    assert not mechanism.is_satisfying({**c, 'irange': -1})
    assert not mechanism.is_satisfying({**c, 'irange': 0})
    assert not mechanism.is_satisfying({**c, 'irange': (-1, 0)})
    assert mechanism.is_satisfying({**c, 'irange': (0, 1)})
    assert mechanism.is_satisfying({**c, 'irange': [0, 1]})
    assert mechanism.is_satisfying({**c, 'irange': 1})
    assert mechanism.is_satisfying({**c, 'irange': (0, 2)})
    assert mechanism.is_satisfying({**c, 'irange': (2, 4)})
    assert mechanism.is_satisfying({**c, 'irange': (4, 5)})
    assert mechanism.is_satisfying({**c, 'irange': (4, 6)})
    assert not mechanism.is_satisfying({**c, 'irange': (6, 7)})

    assert not mechanism.is_satisfying({**c, 'frange': (-1, 0)})
    assert mechanism.is_satisfying({**c, 'frange': (0, 1)})
    assert mechanism.is_satisfying({**c, 'frange': (0, 2)})
    assert mechanism.is_satisfying({**c, 'frange': (2, 4)})
    assert mechanism.is_satisfying({**c, 'frange': (4, 5)})
    assert mechanism.is_satisfying({**c, 'frange': (4, 6)})
    assert not mechanism.is_satisfying({**c, 'frange': (6, 7)})

    # list
    assert not mechanism.is_satisfying({**c, 'ilist': (-1, 0)})
    assert  mechanism.is_satisfying({**c, 'ilist': (0, 1)})
    assert  mechanism.is_satisfying({**c, 'ilist': (3, 10)})

    assert not mechanism.is_satisfying({**c, 'ilist': [-1, 0]})
    assert  mechanism.is_satisfying({**c, 'ilist': [0, 1]})
    assert  not mechanism.is_satisfying({**c, 'ilist': [3, 10]})

    assert not mechanism.is_satisfying({**c, 'ilist': 0})
    assert mechanism.is_satisfying({**c, 'ilist': 9})


def test_can_accept_more_agents(mechanism):
    mechanism.max_n_agents = 2
    assert mechanism.can_accept_more_agents() is True
    mechanism.add(RandomNegotiator(mechanism.outcomes), ufun=lambda x: 5.0)
    assert mechanism.can_accept_more_agents() is True
    mechanism.add(RandomNegotiator(mechanism.outcomes), ufun=lambda x: 5.0)
    assert mechanism.can_accept_more_agents() is False


def test_dynamic_entry(static_mechanism: Mechanism):
    a = RandomNegotiator(static_mechanism.outcomes)
    assert static_mechanism.can_enter(a)
    assert not static_mechanism.can_leave(a)

    static_mechanism.add(a, ufun=lambda x: 5.0)
    static_mechanism.add(RandomNegotiator(static_mechanism.outcomes), ufun=lambda x: 5.0)


def test_mechanism_fails_on_less_than_two_agents(static_mechanism):
    assert not static_mechanism.run().started
    static_mechanism.add(RandomNegotiator(static_mechanism.outcomes), ufun=lambda x: 5.0)
    assert not static_mechanism.run().started
    static_mechanism.add(RandomNegotiator(static_mechanism.outcomes), ufun=lambda x: 5.0)
    static_mechanism.run()
    assert len(static_mechanism.history) > 0 and static_mechanism.state.broken is False


def test_mechanism_fails_on_less_than_two_agents_dynamic(dynamic_mechanism):
    assert not dynamic_mechanism.run().broken
    dynamic_mechanism.add(RandomNegotiator(dynamic_mechanism.outcomes), ufun=lambda x: 5.0)
    assert not dynamic_mechanism.run().broken
    dynamic_mechanism.add(RandomNegotiator(dynamic_mechanism.outcomes), ufun=lambda x: 5.0)
    dynamic_mechanism.run()
    assert len(dynamic_mechanism.history) > 0 and dynamic_mechanism.state.broken is False


def test_mechanisms_get_some_rounds():
    lengths = []
    for _ in range(10):
        p = MyMechanism(dynamic_entry=False)
        p.add(RandomNegotiator(outcomes=10), ufun=lambda x: 5.0)
        p.add(RandomNegotiator(outcomes=10), ufun=lambda x: 5.0)
        p.run()
        lengths.append(len(p.history))

    assert not all(_ < 2 for _ in lengths)


def test_alternating_offers_mechanism():
    p = SAOMechanism(outcomes=10, n_steps=10, dynamic_entry=False, publish_n_acceptances=True, publish_proposer=True)
    to_be_offered = [(0,), (1,), (2,)]
    to_be_accepted = [(2,)]
    a1 = LimitedOutcomesNegotiator(acceptable_outcomes=to_be_offered, outcomes=10)
    a2 = LimitedOutcomesAcceptor(acceptable_outcomes=to_be_accepted, outcomes=10)
    p.add(a1, ufun=MappingUtilityFunction(lambda x: x[0]+1.0))
    p.add(a2, ufun=MappingUtilityFunction(lambda x: x[0] + 1.0))
    p.run()
    a1offers = [s.current_offer for s in p.history if s.current_proposer == a1.id]
    a2offers = [s.current_offer for s in p.history if s.current_proposer == a2.id]
    assert len(p.history) > 0
    assert len(a2offers) > 0, 'acceptor did offer'
    assert p.agreement is None or p.agreement in to_be_accepted, 'acceptor accepted the correct offer'
    assert all([_ in to_be_offered for _ in a1offers])


def test_alternating_offers_mechanism_fails_on_no_offerer():
    p = SAOMechanism(outcomes=10, n_steps=10, dynamic_entry=False, publish_n_acceptances=True, publish_proposer=True)
    to_be_offered = [(0,), (1,), (2,)]
    to_be_accepted = [(2,)]
    a1 = LimitedOutcomesAcceptor(acceptable_outcomes=to_be_offered, outcomes=10)
    a2 = LimitedOutcomesAcceptor(acceptable_outcomes=to_be_accepted, outcomes=10)
    p.add(a1, ufun=MappingUtilityFunction(lambda x: x[0]+1.0))
    p.add(a2, ufun=MappingUtilityFunction(lambda x: x[0] + 1.0))
    try:
        p.run()
    except RuntimeError:
        pass
    a1offers = [s.current_offer for s in p.history if s.current_proposer is a1.id]
    a2offers = [s.current_offer for s in p.history if s.current_proposer is a2.id]
    assert len(p.history) > 0
    assert len(a2offers) == 0, 'acceptor did not offer'
    assert len(a1offers) == 0, 'acceptor did not offer'
    assert p.agreement is None, 'no agreement'


def test_alternating_offers_mechanism_with_one_agent_run():
    n_outcomes, n_steps = 10, 10
    accepted = [(2,), (3,), (4,), (5,)]
    neg = SAOMechanism(outcomes=n_outcomes, n_steps=n_steps)
    opponent = LimitedOutcomesNegotiator(outcomes=n_outcomes,
                                         acceptable_outcomes=accepted,
                                         acceptance_probabilities=[1.0] * len(accepted))
    neg.add(opponent)
    neg.add(opponent)
    assert neg.pareto_frontier(sort_by_welfare=True)[0] == [(1.0,)]

    neg.run()


def test_same_utility_leads_to_agreement():
    n_outcomes, n_steps = 10, 10
    accepted = [(2,), (3,), (4,), (5,)]
    neg = SAOMechanism(outcomes=n_outcomes, n_steps=n_steps)
    opponent = LimitedOutcomesNegotiator(outcomes=n_outcomes,
                                         acceptable_outcomes=accepted,
                                         acceptance_probabilities=[1.0] * len(accepted))
    acceptor = LimitedOutcomesAcceptor(outcomes=n_outcomes,
                                       acceptable_outcomes=accepted,
                                       acceptance_probabilities=[1.0] * len(accepted))
    neg.add(opponent)
    neg.add(acceptor)
    assert neg.pareto_frontier(sort_by_welfare=True)[0] == [(1.0, 1.0)]

    state = neg.run()
    assert state.agreement is not None
    assert state.step < 4


if __name__ == '__main__':
    pytest.main(args=[__file__])



"""Implements all builtin banks."""
from abc import ABC
from collections import defaultdict
from typing import Dict, List, Optional, TYPE_CHECKING, Any

from negmas import Issue, Negotiator, Mechanism, AgentMechanismInterface, MechanismState
from negmas.situated import Agent, RenegotiationRequest, Contract, Breach
from .common import *

if TYPE_CHECKING:
    from .agent import SCMLAgent

__all__ = [
    'DefaultBank', 'Bank'
]


class Bank(Agent, ABC):
    """Base class for all banks"""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._world = None

    def _respond_to_negotiation_request(self, initiator: str, partners: List[str], issues: List[Issue],
                                        annotation: Dict[str, Any], mechanism: AgentMechanismInterface,
                                        role: Optional[str], req_id: Optional[str]) -> Optional[Negotiator]:
        pass

    def on_neg_request_rejected(self, req_id: str, by: Optional[List[str]]):
        pass

    def on_neg_request_accepted(self, req_id: str, mechanism: AgentMechanismInterface):
        pass

    def on_negotiation_failure(self, partners: List[str], annotation: Dict[str, Any],
                               mechanism: AgentMechanismInterface, state: MechanismState) -> None:
        pass

    def on_negotiation_success(self, contract: Contract, mechanism: AgentMechanismInterface) -> None:
        pass

    def on_contract_signed(self, contract: Contract) -> None:
        pass

    def on_contract_cancelled(self, contract: Contract, rejectors: List[str]) -> None:
        pass

    def sign_contract(self, contract: Contract) -> Optional[str]:
        pass

    def respond_to_negotiation_request(self, initiator: str, partners: List[str], issues: List[Issue],
                                       annotation: Dict[str, Any], mechanism: Mechanism, role: Optional[str],
                                       req_id: str) -> Optional[Negotiator]:
        pass


class DefaultBank(Bank):
    """Represents a bank in the world"""

    def init(self):
        pass

    def respond_to_negotiation_request(self, initiator: str, partners: List[str], issues: List[Issue],
                                       annotation: Dict[str, Any], mechanism: Mechanism, role: Optional[str],
                                       req_id: str) -> Optional[Negotiator]:
        pass

    def __init__(self, minimum_balance: float, interest_rate: float
                 , interest_max: float, balance_at_max_interest: float, installment_interest: float
                 , time_increment: float, a2f: Dict[str, Factory]
                 , disabled: bool = False
                 , name: str = None):
        super().__init__(name=name)
        self.storage: Dict[int, int] = defaultdict(int)
        self.wallet: float = 0.0
        self.disabled = disabled
        self.loans: Dict[SCMLAgent, List[Loan]] = defaultdict(list)
        self.minimum_balance = minimum_balance
        self.interest_rate = interest_rate
        self.interest_max = interest_max
        self.installment_interest = installment_interest
        self.time_increment = time_increment
        self.balance_at_max_interest = balance_at_max_interest
        self._credit_rating: Dict[str, float] = defaultdict(float)
        self.a2f = a2f

    def set_renegotiation_agenda(self, contract: Contract, breaches: List[Breach]) -> Optional[RenegotiationRequest]:
        return None

    def respond_to_renegotiation_request(self, contract: Contract, breaches: List[Breach]
                                         , agenda: RenegotiationRequest) -> Optional[Negotiator]:
        raise ValueError('The bank does not receive callbacks')

    def _evaluate_loan(self, agent: 'SCMLAgent', amount: float, n_installments: int, starts_at: int
                       , installment_loan=False) -> Optional[Loan]:
        """Evaluates the interest that will be imposed on the agent to buy_loan that amount"""
        if self.disabled:
            return None
        factory = self.a2f[agent.id]
        balance = factory.balance

        if self.interest_rate is None:
            return None
        interest = self.installment_interest if installment_loan else self.interest_rate

        if balance < 0 and self.interest_max is not None:
            interest += balance * (interest - self.interest_max) / self.balance_at_max_interest
        interest += max(0, starts_at - self.awi.current_step) * self.time_increment
        interest += self._credit_rating[agent.id]
        total = amount * (1 + interest) ** n_installments
        installment = total / n_installments
        if self.minimum_balance is not None and balance - total < - self.minimum_balance:
            return None
        return Loan(amount=amount, total=total, interest=interest
                    , n_installments=n_installments, installment=installment, starts_at=starts_at)

    def evaluate_loan(self, agent: 'SCMLAgent', amount: float, start_at: int, n_installments: int) -> Optional[Loan]:
        """Evaluates the interest that will be imposed on the agent to buy_loan that amount"""
        if self.disabled:
            return None
        return self._evaluate_loan(agent=agent, amount=amount, n_installments=n_installments, installment_loan=False
                                   , starts_at=start_at)

    def _buy_loan(self, agent: 'SCMLAgent', loan: Loan, beneficiary: Agent
                  , contract: Optional[Contract], bankrupt_if_rejected=False) -> Optional[Loan]:
        if self.disabled:
            return None
        if loan is None:
            return loan
        factory = self.a2f[agent.id]
        if agent.confirm_loan(loan=loan, bankrupt_if_rejected=bankrupt_if_rejected):
            self.loans[agent].append(loan)
            self.awi.logdebug(f'Bank: {agent.name} borrowed {str(loan)}')
            factory.receive(loan.amount)
            factory.add_loan(loan.total)
            self.wallet -= loan.amount
            return loan
        elif bankrupt_if_rejected:
            # the agent rejected a loan with bankrupt_if_rejected, bankrupt the agent
            self._world.make_bankrupt(agent, amount=loan.amount, beneficiary=beneficiary, contract=contract)
            return None
        return None

    def buy_loan(self, agent: 'SCMLAgent', amount: float, n_installments: int
                 , beneficiary: Agent, contract: Optional[Contract], force: bool = False) -> Optional[Loan]:
        """Gives a loan of amount to agent at the interest calculated using `evaluate_loan`"""
        if self.disabled:
            return None
        loan = self.evaluate_loan(amount=amount, agent=agent, n_installments=n_installments
                                  , start_at=self.awi.current_step)
        return self._buy_loan(agent=agent, loan=loan, bankrupt_if_rejected=force, beneficiary=beneficiary
                              , contract=contract)

    def step(self):
        """Takes payments from agents"""
        # apply interests and pay loans
        # -----------------------------
        if self.disabled:
            return
        t = self.awi.current_step
        delayed_payments = 0.0
        # for every agent with loans
        for agent, loans in self.loans.items():
            factory = self.a2f[agent.id]
            keep = [True] * len(loans)  # a flag to tell whether a loan is to be kept for future processing
            unpaid = []  # any new loans that may arise from failure to pay installments
            unavailable = 0.0
            for i, loan in enumerate(loans):
                # if there are no remaining installments or I am in the grace period do not do anything
                if loan.n_installments <= 0 or loan.starts_at > t:
                    continue

                # pay as much as possible from the agent's wallet (which may be zero)
                wallet = factory.wallet
                payment = max(0.0, min(loan.installment, wallet))
                loan.amount -= payment
                factory.pay(payment)
                factory.add_loan(-payment)
                self.wallet += payment

                if payment >= loan.installment:
                    # reduce the number of remaining installments if needed
                    loan.n_installments -= 1
                else:
                    # if the payment is not enough for the installment, try to get a new loan
                    unavailable += loan.installment - payment
                    unpaid.append((i, loan))
                # if the loan is completely paid, mark it for removal
                if loan.n_installments <= 0:
                    keep[i] = False
                self.awi.logdebug(f'Bank: {agent.name} payed {payment} (of {loan.installment}) '
                                  f'[{loan.n_installments} remain]')
            if unavailable > 0.0:
                new_loan = self._evaluate_loan(agent=agent, amount=unavailable, n_installments=1
                                               , installment_loan=True, starts_at=t + 1)
                if new_loan is None:
                    self._reduce_credit_rating(agent=agent, unavailable=unavailable)
                    self.awi.logdebug(f'Bank: CR of {agent.name} was reduced for failure to pay {unavailable}')
                elif self._buy_loan(agent=agent, loan=new_loan, bankrupt_if_rejected=True, beneficiary=self
                                    , contract=None) is not None:
                    self.awi.logdebug(f'Bank: {agent.name} payed an installment by a new loan {str(new_loan)}')
                    factory.add_loan(-new_loan.amount)
                    factory.pay(new_loan.amount)
                    self.wallet += new_loan.amount
                    for indx, loan in unpaid:
                        loan.amount -= loan.installment
                        loan.n_installments -= 1
                        if loan.n_installments <= 0:
                            keep[indx] = False

            # remove marked loans (that were completely paid)
            self.loans[agent] = [l for i, l in enumerate(loans) if keep[i]]

        # remove records of agents that paid all their loans
        self.loans = {k: v for k, v in self.loans.items() if len(v) > 0}

    def _reduce_credit_rating(self, agent: Agent, unavailable: float):
        """Updates the credit rating when the agent fails to pay an installment"""
        self._credit_rating[agent.id] -= unavailable

    def credit_rating(self, agent_id: str) -> float:
        if self.disabled:
            return 1
        return self._credit_rating.get(agent_id, 1.0)



from abc import ABC
from collections import defaultdict
from typing import Dict, Tuple, List, Optional, TYPE_CHECKING, Any

from negmas import Mechanism, AgentMechanismInterface, MechanismState
from negmas.negotiators import Negotiator
from negmas.outcomes import Issue
from negmas.situated import Agent, RenegotiationRequest, Breach, Contract
from .agent import SCMLAgent
from .common import InsurancePolicy, Factory

if TYPE_CHECKING:
    from .world import SCMLWorld

__all__ = [
    'DefaultInsuranceCompany',
    'InsuranceCompany'
]


class InsuranceCompany(Agent, ABC):
    """Base class for all insurance companies"""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._world: Optional[SCMLWorld] = None

    def _respond_to_negotiation_request(self, initiator: str, partners: List[str], issues: List[Issue],
                                        annotation: Dict[str, Any], mechanism: AgentMechanismInterface,
                                        role: Optional[str], req_id: Optional[str]) -> Optional[Negotiator]:
        pass

    def on_neg_request_rejected(self, req_id: str, by: Optional[List[str]]):
        pass

    def on_neg_request_accepted(self, req_id: str, mechanism: AgentMechanismInterface):
        pass

    def on_negotiation_failure(self, partners: List[str], annotation: Dict[str, Any],
                               mechanism: AgentMechanismInterface, state: MechanismState) -> None:
        pass

    def on_negotiation_success(self, contract: Contract, mechanism: AgentMechanismInterface) -> None:
        pass

    def on_contract_signed(self, contract: Contract) -> None:
        pass

    def on_contract_cancelled(self, contract: Contract, rejectors: List[str]) -> None:
        pass

    def sign_contract(self, contract: Contract) -> Optional[str]:
        pass

    def respond_to_negotiation_request(self, initiator: str, partners: List[str], issues: List[Issue],
                                       annotation: Dict[str, Any], mechanism: Mechanism, role: Optional[str],
                                       req_id: str) -> Optional[Negotiator]:
        pass


class DefaultInsuranceCompany(InsuranceCompany):
    """Represents an insurance company in the world"""

    def __init__(self, premium: float, premium_breach_increment: float, premium_time_increment: float
                 , a2f: Dict[str, Factory], disabled=False, name: str = None):
        super().__init__(name=name)
        self.premium_breach_increment = premium_breach_increment
        self.premium = premium
        self.disabled = disabled
        self.premium_time_increment = premium_time_increment
        self.insured_contracts: Dict[Tuple[Contract, str], InsurancePolicy] = dict()
        self.storage: Dict[int, int] = defaultdict(int)
        self.wallet: float = 0.0
        self.a2f = a2f

    def init(self):
        pass

    def set_renegotiation_agenda(self, contract: Contract, breaches: List[Breach]) -> Optional[RenegotiationRequest]:
        return None

    def respond_to_renegotiation_request(self, contract: Contract, breaches: List[Breach]
                                         , agenda: RenegotiationRequest) -> Optional[Negotiator]:
        raise ValueError('The insurance company does not receive callbacks')

    def evaluate_insurance(self, contract: Contract, insured: SCMLAgent, against: SCMLAgent
                           , t: int = None) -> Optional[float]:
        """Can be called to evaluate the premium for insuring the given contract against breaches committed by others

        Args:

            against: The `SCMLAgent` to insure against
            contract: hypothetical contract
            insured: The `SCMLAgent` to buy the insurance
            t: time at which the policy is to be bought. If None, it means current step
        """
        if self.disabled:
            return None
        # fail if no premium
        if self.premium is None:
            return None

        # assume the insurance is to be bought now if needed
        if t is None:
            t = self.awi.current_step

        # find the delay from contract signing. The more this is the more expensive the insurance will be
        if contract.signed_at is None:
            dt = 0
        else:
            dt = max(0, t - contract.signed_at)

        # fail if the insurance is to be bought at or after the agreed upon delivery time
        if t >= contract.agreement.get('time', -1):
            return None

        # find the total breach of the agent I am insuring against. The more this is, the more expensive the insurance
        breaches = self.awi.bb_query(section='breaches', query={'perpetrator': against})
        b = 0
        if breaches is not None:
            for _, breach in breaches.items():
                b += breach.level
        return (self.premium + b * self.premium_breach_increment) * (1 + self.premium_time_increment * dt)

    def buy_insurance(self, contract: Contract, insured: SCMLAgent, against: SCMLAgent) -> Optional[InsurancePolicy]:
        """Buys insurance for the contract by the premium calculated by the insurance company.

        Remarks:
            The agent can call `evaluate_insurance` to find the premium that will be used.
        """
        if self.disabled:
            return None
        premium = self.evaluate_insurance(contract=contract, t=self.awi.current_step, insured=insured, against=against)
        factory = self.a2f[insured.id]
        if premium is None or factory.wallet < premium:
            return None
        factory.pay(premium)
        self.wallet += premium
        policy = InsurancePolicy(contract=contract, at_time=self.awi.current_step, against=against, premium=premium)
        self.insured_contracts[(contract, against.id)] = policy
        return policy

    def pay_insurance(self, contract: Contract, perpetrator: SCMLAgent) -> bool:
        """

        Args:
            contract:
            perpetrator:

        Returns:

        """
        if self.disabled:
            return False
        if (contract, perpetrator.id) in self.insured_contracts.keys():
            del self.insured_contracts[(contract, perpetrator.id)]
            return True
        return False

    def step(self):
        """does nothing"""



#!/usr/bin/env python
"""The NegMAS universal command line tool"""

import os
import traceback
from functools import partial
from pathlib import Path
from pprint import pformat
from time import perf_counter

import click
import pandas as pd
import progressbar
import yaml
from tabulate import tabulate

import negmas
from negmas import save_stats, tournaments
from negmas.apps.scml import *
from negmas.apps.scml.utils import anac2019_world, balance_calculator
from negmas.helpers import humanize_time, unique_name
from negmas.java import init_jnegmas_bridge, jnegmas_bridge_is_running

try:
    # disable a warning in yaml 1b1 version
    yaml.warnings({'YAMLLoadWarning': False})
except:
    pass

n_completed = 0
n_total = 0


def print_progress(_, i, n) -> None:
    """Prints the progress of a tournament"""
    global n_completed, n_total
    n_completed = i + 1
    n_total = n
    print(f'{n_completed:04} of {n:04} worlds completed ({n_completed / n:0.2%})', flush=True)


def print_world_progress(world) -> None:
    """Prints the progress of a world"""
    step = world.current_step + 1
    s = f'World# {n_completed:04}: {step:04}  of {world.n_steps:04} ' \
        f'steps completed ({step / world.n_steps:0.2f}) '
    if n_total > 0:
        s += f'TOTAL: ({n_completed + step / world.n_steps / n_total:0.2f})'
    print(s, flush=True)


click.option = partial(click.option, show_default=True)


@click.group()
def cli():
    pass


@cli.command(help='Run a tournament between candidate agent types')
@click.option('--name', '-n', default='random',
              help='The name of the tournament. The special value "random" will result in a random name')
@click.option('--steps', '-s', default=60, help='Number of steps.')
@click.option('--ttype', '--tournament-type', '--tournament', default='anac2019'
    , help='The config to use. Default is ANAC 2019')
@click.option('--timeout', '-t', default=0, help='Timeout after the given number of seconds (0 for infinite)')
@click.option('--runs', default=5, help='Number of runs for each configuration')
@click.option('--max-runs', default=-1, help='Maximum total number of runs. Zero or negative numbers mean no limit')
@click.option('--randomize/--permutations', default=False, help='Random worlds or try all permutations up to max-runs')
@click.option('--competitors'
    , default='negmas.apps.scml.DoNothingFactoryManager;negmas.apps.scml.GreedyFactoryManager'
    , help='A semicolon (;) separated list of agent types to use for the competition.')
@click.option('--jcompetitors', '--java-competitors'
    , default=''
    , help='A semicolon (;) separated list of agent types to use for the competition.')
@click.option('--parallel/--serial', default=True, help='Run a parallel/serial tournament on a single machine')
@click.option('--distributed/--single-machine', default=False, help='Run a distributed tournament using dask')
@click.option('--log', '-l', default='~/negmas/logs/tournaments',
              help='Default location to save logs (A folder will be created under it)')
@click.option('--verbosity', default=1, help='verbosity level (from 0 == silent to 1 == world progress)')
@click.option('--configs-only/--run', default=False, help='configs_only')
@click.option('--reveal-names/--hidden-names', default=False, help='Reveal agent names (should be used only for '
                                                                   'debugging)')
@click.option('--ip', default='127.0.0.1', help='The IP address for a dask scheduler to run the distributed tournament.'
                                                ' Effective only if --distributed')
@click.option('--port', default=8786, help='The IP port number a dask scheduler to run the distributed tournament.'
                                           ' Effective only if --distributed')
def tournament(name, steps, parallel, distributed, ttype, timeout, log, verbosity, configs_only,
               reveal_names, ip, port, runs, max_runs, randomize, competitors, jcompetitors):
    if timeout <= 0:
        timeout = None
    if name == 'random':
        name = None
    if max_runs <= 0:
        max_runs = None
    parallelism = 'distributed' if distributed else 'parallel' if parallel else 'serial'
    start = perf_counter()
    all_competitors = competitors.split(';')
    all_competitors_params = [dict() for _ in range(len(all_competitors))]
    if jcompetitors is not None and len(jcompetitors) > 0:
        jcompetitor_params = [{'java_class_name':_} for _ in jcompetitors.split(';')]
        jcompetitors = ['negmas.apps.scml.JavaFactoryManager'] * len(jcompetitor_params)
        all_competitors += jcompetitors
        all_competitors_params += jcompetitor_params
        print('You are using some Java agents. The tournament MUST run serially')
        parallelism = 'serial'
        if not jnegmas_bridge_is_running():
            print('Error: You are using java competitors but jnegmas bridge is not running\n\nTo correct this issue'
                  ' run the following command IN A DIFFERENT TERMINAL because it will block:\n\n'
                  '$ negmas jnegmas')
            exit(1)
    if ttype.lower() == 'anac2019':
        results = tournaments.tournament(competitors=all_competitors
                                         , competitor_params=all_competitors_params
                                         , agent_names_reveal_type=reveal_names
                                         , tournament_path=log, total_timeout=timeout
                                         , parallelism=parallelism, scheduler_ip=ip, scheduler_port=port
                                         ,
                                         world_progress_callback=print_world_progress if verbosity > 1 and not distributed else None
                                         , name=name, verbose=verbosity > 0, n_runs_per_config=runs,
                                         max_n_configs=max_runs
                                         , world_generator=anac2019_world, score_calculator=balance_calculator
                                         , configs_only=configs_only, randomize=randomize
                                         , n_steps=steps)
    else:
        print('Only anac2019 tournament type is supported')
        exit(1)
    if configs_only:
        print(f'Saved all configs to {str(results)}')
        print(f'Finished in {humanize_time(perf_counter() - start)} [config-only]')
        return
    print(tabulate(results.total_scores, headers='keys', tablefmt='psql'))
    print(f'Finished in {humanize_time(perf_counter() - start)} [{parallelism}]')


@cli.command(help='Run an SCML world simulation')
@click.option('--steps', default=120, help='Number of steps.')
@click.option('--levels', default=3, help='Number of intermediate production levels (processes). '
                                          '-1 means a single product and no factories.')
@click.option('--neg-speedup', default=21, help='Negotiation Speedup.')
@click.option('--negotiator', default='negmas.sao.AspirationNegotiator',
              help='Negotiator type to use for builtin agents.')
@click.option('--min-consumption', default=3, help='The minimum number of units consumed by each consumer at every '
                                                   'time-step.')
@click.option('--max-consumption', default=5, help='The maximum number of units consumed by each consumer at every '
                                                   'time-step.')
@click.option('--agents', default=5, help='Number of agents (miners/negmas.consumers) per production level')
@click.option('--horizon', default=20, help='Consumption horizon.')
@click.option('--transport', default=0, help='Transportation Delay.')
@click.option('--time', default=60 * 90, help='Total time limit.')
@click.option('--neg-time', default=60 * 4, help='Time limit per single negotiation')
@click.option('--neg-steps', default=20, help='Number of rounds per single negotiation')
@click.option('--sign', default=1, help='The default delay between contract conclusion and signing')
@click.option('--guaranteed', default=False, help='Whether to only sign contracts that are guaranteed not to cause '
                                                  'breaches')
@click.option('--lines', default=10, help='The number of lines per factory')
@click.option('--retrials', default=5, help='The number of times an agent re-tries on failed negotiations')
@click.option('--use-consumer', default=True, help='Use internal consumer object in factory managers')
@click.option('--max-insurance', default=100, help='Use insurance against partner in factory managers up to this '
                                                   'premium')
@click.option('--riskiness', default=0.0, help='How risky is the default factory manager')
@click.option('--log', default='~/negmas/logs',
              help='Default location to save logs (A folder will be created under it)')
def scml(steps, levels, neg_speedup, negotiator, agents, horizon, min_consumption, max_consumption
         , transport, time, neg_time
         , neg_steps, sign, guaranteed, lines, retrials, use_consumer, max_insurance, riskiness, log):
    params = {
        "steps": steps,
        "levels": levels,
        "neg_speedup": neg_speedup,
        "negotiator": negotiator,
        "agents": agents,
        "horizon": horizon,
        "min_consumption": min_consumption,
        "max_consumption": max_consumption,
        "transport": transport,
        "time": time,
        "neg_time": neg_time,
        "neg_steps": neg_steps,
        "sign": sign,
        "guaranteed": guaranteed,
        "lines": lines,
        "retrials": retrials,
        "use_consumer": use_consumer,
        "max_insurance": max_insurance,
        "riskiness": riskiness
    }
    neg_speedup = neg_speedup if neg_speedup is not None and neg_speedup > 0 else None
    if min_consumption == max_consumption:
        consumption = min_consumption
    else:
        consumption = (min_consumption, max_consumption)
    customer_kwargs = {'negotiator_type': negotiator, 'consumption_horizon': horizon}
    miner_kwargs = {'negotiator_type': negotiator, 'n_retrials': retrials}
    factory_kwargs = {'negotiator_type': negotiator, 'n_retrials': retrials
        , 'sign_only_guaranteed_contracts': guaranteed, 'use_consumer': use_consumer
        , 'riskiness': riskiness, 'max_insurance_premium': max_insurance}
    if log.startswith('~/'):
        log_dir = Path.home() / log[2:]
    else:
        log_dir = Path(log)
    log_dir = log_dir / unique_name(base='scml', add_time=True, rand_digits=0)
    log_dir = log_dir.absolute()
    os.makedirs(log_dir, exist_ok=True)
    log_file_name = str(log_dir / 'log.txt')
    stats_file_name = str(log_dir / 'stats.json')
    params_file_name = str(log_dir / 'params.json')
    exception = None
    world = SCMLWorld.chain_world(log_file_name=log_file_name, n_steps=steps
                                  , negotiation_speed=neg_speedup
                                  , n_intermediate_levels=levels
                                  , n_miners=agents
                                  , n_consumers=agents
                                  , n_factories_per_level=agents
                                  , consumption=consumption
                                  , consumer_kwargs=customer_kwargs
                                  , miner_kwargs=miner_kwargs
                                  , default_manager_params=factory_kwargs
                                  , transportation_delay=transport, time_limit=time, neg_time_limit=neg_time
                                  , neg_n_steps=neg_steps, default_signing_delay=sign
                                  , n_lines_per_factory=lines)
    failed = False
    strt = perf_counter()
    try:
        for i in progressbar.progressbar(range(world.n_steps), max_value=world.n_steps):
            elapsed = perf_counter() - strt
            if world.time_limit is not None and elapsed >= world.time_limit:
                break
            if not world.step():
                break
    except Exception:
        exception = traceback.format_exc()
        failed = True
    elapsed = perf_counter() - strt

    def print_and_log(s):
        world.logdebug(s)
        print(s)

    world.logdebug(f'{pformat(world.stats, compact=True)}')
    world.logdebug(f'=================================================\n'
                   f'steps: {steps}, horizon: {horizon}, time: {time}, levels: {levels}, agents_per_level: '
                   f'{agents}, lines: {lines}, guaranteed: {guaranteed}, negotiator: {negotiator}\n'
                   f'consumption: {consumption}'
                   f', transport_to: {transport}, sign: {sign}, speedup: {neg_speedup}, neg_steps: {neg_steps}'
                   f', retrials: {retrials}'
                   f', neg_time: {neg_time}\n'
                   f'==================================================')

    save_stats(world=world, log_dir=log_dir, params=params)

    if len(world.saved_contracts) > 0:
        data = pd.DataFrame(world.saved_contracts)
        data = data.sort_values(['delivery_time'])
        data = data.loc[:, ['seller_type', 'buyer_type', 'seller_name', 'buyer_name', 'delivery_time', 'unit_price'
                               , 'quantity', 'product_name', 'n_neg_steps', 'signed_at', 'concluded_at', 'nullified_at'
                               , 'cfp']]
        print_and_log(tabulate(data, headers='keys', tablefmt='psql'))
        n_executed = sum(world.stats['n_contracts_executed'])
        n_negs = sum(world.stats["n_negotiations"])
        n_contracts = len(world.saved_contracts)
        winners = [f'{_.name} gaining {world.a2f[_.id].balance / world.a2f[_.id].initial_balance - 1.0:0.0%}'
                   for _ in world.winners]
        print_and_log(f'{n_contracts} contracts :-) [N. Negotiations: {n_negs}'
                      f', Agreement Rate: {world.agreement_rate:0.0%}]')
        print_and_log(f'Executed: {world.contract_execution_fraction:0.0%}'
                      f', Breached: {world.breach_rate:0.0%}'
                      f', N. Executed: {n_executed}'
                      f', Business size: {world.business_size}\n'
                      f'Winners: {winners}\n'
                      f'Running Time {humanize_time(elapsed)}')
    else:
        print_and_log('No contracts! :-(')
        print_and_log(f'Running Time {humanize_time(elapsed)}')

    if failed:
        print(exception)
        world.logdebug(exception)
        print(f'FAILED at step {world.current_step} of {world.n_steps}\n')


@cli.command(help='Start the bridge to genius (to use GeniusNegotiator)')
@click.option('--path', '-p', default='auto', help='Path to genius-8.0.4.jar with embedded NegLoader. Use "auto" to '
                                                   'read the path from ~/negmas/config.json'
                                                   '\n\tConfig key is genius_bridge_jar'
                                                   '\nYou can download this jar from: '
                                                   'http://www.yasserm.com/scml/genius-8.0.4-bridge.jar')
@click.option('--port', '-r', default=0, help='Port to run the NegLoader on. Pass 0 for the default value')
@click.option('--force/--no-force', default=False, help='Force trial even if an earlier instance exists')
def genius(path, port, force):
    negmas.init_genius_bridge(path=path if path != 'auto' else None, port=port, force=force)


@cli.command(help='Start the bridge to JNegMAS (to use Java agents in worlds)')
@click.option('--path', '-p', default='auto', help='Path to jnegmas*.jar with. Use "auto" to '
                                                   'read the path from ~/negmas/config.json.'
                                                   '\n\tConfig key is jnegmas_jar'
                                                   '\nYou can download the latest version of this jar from: '
                                                   'http://www.yasserm.com/scml/jnegmas-all.jar')
@click.option('--port', '-r', default=0, help='Port to run the jnegmas on. Pass 0 for the default value')
def jnegmas(path, port):
    init_jnegmas_bridge(path=path if path != 'auto' else None, port=port)
    input('Press C^c to quit')


if __name__ == '__main__':
    cli()
